@@ -21,7 +21,11 @@
 #ifndef RBFFI_POINTER_H
 #define	RBFFI_POINTER_H
 
-#include <stdbool.h>
+#ifndef _MSC_VER
+# include <stdbool.h>
+#else
+# include "win32/stdbool.h"
+#endif
 
 #ifdef	__cplusplus
 extern "C" {
